#!/usr/bin/env python
# ======================================================================
# Copyright (C) 2008 and onwards  Robert Muth <robert at muth dot org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ======================================================================

import fcntl
import os
import struct
import sys

# these constants taken from linux/cdrom.h
CDROMREADTOCHDR = 0x5305
CDROMREADTOCHDR_FORMAT = 'BB'

CDROMREADTOCENTRY = 0x5306
CDROMREADTOCENTRY_FORMAT = 'BBBBBBBB'

CDROM_LEADOUT = 0xAA
CDROM_MSF = 0x02
CDROM_DATA_TRACK = 0x04

def GetFirstAndLastTrack(fd):
    buf = struct.pack(CDROMREADTOCHDR_FORMAT, 0, 0)
    try:
        buf = fcntl.ioctl(fd, CDROMREADTOCHDR, buf)
    except IOError, x:
        if x.errno == 123:
            return None, None
        else:
            raise x
    return struct.unpack(CDROMREADTOCHDR_FORMAT, buf)

def GetTrackInfo(fd, track):
    buf = struct.pack(CDROMREADTOCENTRY_FORMAT,
                      track, 0, CDROM_MSF, 0, 0, 0, 0, 0)
    try:
        buf = fcntl.ioctl(fd, CDROMREADTOCENTRY, buf)
    except IOError, x:
        if x.errno == 123:
            return None, None, None, None, None, None
        else:
            raise x
    result = struct.unpack(CDROMREADTOCENTRY_FORMAT, buf)
    is_data = (result[1] & 0xf0) == CDROM_DATA_TRACK * 16
    track = result[0]
    m = result[4]
    s = result[5]
    f = result[6]
    return track, is_data, m, s, f

def Msf2Pos(m, s, f):
    return (m * 60 + s) * 75 + f

def GetTrackList(fd):
    first, last = GetFirstAndLastTrack(fd)
    if first is None:
        return None
    tracks = []
    for t in range(first, last + 1) + [CDROM_LEADOUT]:
        t, d, m, s, f= GetTrackInfo(fd, t)
        if t is None:
            return None
        tracks.append(Msf2Pos(m, s, f))
    return tracks

def Duration(tl):
    return (tl[-1] - tl[0]) / 75

def TrackListToDiscId(tl):
    duration = Duration(tl)
    tl = tl[:-1]

    def sum_digits(n):
        ret = 0
        while n > 0:
            ret += n % 10
            n = n / 10
        return ret

    s = sum(sum_digits(x / 75) for x in tl)
    id = (s % 0xff) << 24 | duration << 8 | len(tl)
    return "%08x" % id

def GetCddbInfo(device):
    fd = os.open(device, os.O_RDONLY | os.O_NONBLOCK)
    tl = GetTrackList(fd)
    id = TrackListToDiscId(tl)
    os.close(fd)
    return id, tl

def GetCddbInfoString(device):
    id, tl = GetCddbInfo(device)
    duration = Duration(tl)
    tl = [str(x) for x in tl[:-1]]
    return "%s %d %s %d" % (id, len(tl), " ".join(tl), duration)

if __name__ == '__main__':
    print GetCddbInfoString(sys.argv[1])
    sys.exit(0)
