#!/usr/bin/env python

# ======================================================================
# Copyright (C) 2002 and onwards  Robert Muth <robert at muth dot org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; version 3
# of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# ======================================================================

"""
A python program for interacting with the gnudb.org database.
"""
# python imports
import urllib
import urllib2
import socket
import os
import sys
import getopt
import re
import getpass
import logging
# local imports
import discid

USAGE = """
Usage: freedbtool [mode] [option]*

Retrieve or submit information to and from the freedb.org database.
The mode parameter specifies the concrete command.  Options are used
to change the default behavior of commands.  Currently available modes
are: config,lscat,get,merge,submit The model of operation of freedb
tool is centered around 3 files characterizing a CD

id
    contains track length information for a CD,it is compatible with
    the cddb query command.  A seperate tool discid is used to obtain
    this information.

index
    contains genre and checksum information for a CD, it is compatible
    with the freedb read command. This information is obtained by quering
    the freedb.org with the information in id.

toc
    contains the actual freedb.org database record for a CD.  This
    information is obtained by quering the freedb.org with the
    information in index.


Options:

-t filename
    change default toc filename [toc]

-x filename
    change default index filename [index]

-i filename
    change default id filename [id]

-n number
    specify record number in case of multiple matches

-c clientname
    clientname used to when contacting the server [freedbtool]

-h domain
    specify domain part of user email

-u username
    specify username part of user email

-s server
    specify server [http://gnudb.gnudb.org/]

-d device
    specify device holding the cd [/dev/cdrom]
    if device is empty the cd is not queried and
    the hopefully already existing id file is used

-m db-submit-mode
    specify mode for submit command [submit]

-p protocolnumber
    specify protocol number [6]

-e encodings
    specify sequence of encodings used to parse the result returned
    by freedb [utf8,latin1]

-q freedb-id
   specify freedb id explicitly instead of having it extracted
   from the toc file (for update, submit and scan command)

-g freedb-category
   specify freedb category explicitly instead of having it extracted
   from the index file (for update, and submit command)

Command description:

config
    print version and current defaults

motd
    print message of the day from freedb

lscat
    print categories from freedb

get
    get id, index, toc for current disc from freedb

update
    update existing toc from freedb.org

submit
    submit revised toc to freedb.org

merge
    merge genre info from index into toc file

track
    merge/update track info from id file into toc file

scan
    try all know freedb categories for a given

Written by Robert Muth (robert at muth dot org)
See program file for copyright information
"""

# see lscat commmand
_CDDB_GENRES = ["blues",
                "classical",
                "country",
                "data",
                "folk",
                "jazz",
                "misc",
                "newage",
                "reggae",
                "soundtrack"
                ]

# ======================================================================
def version():
    """
    get cvs/svn revision of this file
    """
    (dummy, ver, dummy2) = '$Revision: 806 $'.split()
    return ver

# ======================================================================
def usage():
    print "Freedbtool Version " + version() + "\n" + USAGE
    return -1

# ======================================================================
def Error(*args):
    logging.fatal(*args)
    sys.exit(-1)
# ======================================================================
def ReadMetaData(device, filename):
    data = None
    try:
        data = discid.GetCddbInfoString(device)
    except Exception, err:
        logging.info(str(err))
    if not data:
        Error("could not read disk in %s", device)

    out = file(filename, "w")
    out.write(data)
    out.close()

# ======================================================================
def decode(line, encodings):
    for enc in encodings:
        try:
            return line.decode(enc)
        except:
            pass
    raise ValueError

# ======================================================================
def encode(line):
    assert type(line) == unicode
    return line.encode("utf8")

# ======================================================================
def read_id_file(filename):
    """
    read id file
    all files written by freedbtool should be utf8
    """
    encodings = ["utf8", "latin1"]

    inp = file(filename, "r")
    lines = inp.readlines()
    inp.close()

    tracks = decode(lines[0], encodings).split()
    return tracks

# ======================================================================
def read_index_file(filename):
    """
    read index file
    all files written by freedbtool should be utf8
    """
    encodings = ["utf8", "latin1"]
    inp = file(filename, "r")
    lines = inp.readlines()
    inp.close()

    if len(lines) > 1:
        logging.warning("multiple lines in index\n")

    disc = decode(lines[0], encodings).split()

    return disc

# ======================================================================
_PATTERN_DISC_ID = re.compile(r'^DISCID\=(.*)$')

def read_toc_file(filename):
    """
    read toc file
    all files written by freedbtool should be utf8
    """
    encodings = ["utf8", "latin1"]
    discids = None

    inp = file(filename, "r")
    lines = [decode(line, encodings) for line in  inp.readlines()]
    inp.close()

    for line in lines:
        match = _PATTERN_DISC_ID.search(line)
        if match:
            discids  = match.group(1).split(",")
            for i in range(len(discids)):
                discids[i] = discids[i].strip()
    return (lines, discids)

# ======================================================================
_PATTERN_REVISION = re.compile(r'^# Revision:\s+([0-9]+)\s*$')

def increment_revision(lines):
    """
    update revision comment if present
    """
    for i in range(len(lines)):
        match = _PATTERN_REVISION.search(lines[i])
        if match:
            revision = int(match.group(1))
            logging.info("found revision %d" % revision)
            lines[i] = u"# Revision: %d\r\n" % (revision+1)
            return
    if revision < 0:
        logging.warning("no revision found")
    return

# ======================================================================
_PatternSubmitted = re.compile(r'^# Submitted via:(.*)$')

def replace_submiter(Config, lines):
    """
    update submitted comment if present
    """
    for i in range(len(lines)):
        match = _PatternSubmitted.search(lines[i])
        if match:
            logging.info("found submitter %s" % match.group(1))
            lines[i] = u"# Submitted via: freedbtool %s\r\n" % Config['version']
            return
    return

# ======================================================================

def backup_file(filename):
    """
    save backups by appending ".1-9" and rotating
    """
    for i in range(8, 0, -1):
        try:
            src = filename + (".%d" % i)
            dst = filename + (".%d" % (i+1))
            #Info("Renaming %s -> %s" % (src,dst))
            os.rename(src, dst)
        except:
            pass
    try:
        os.rename(filename, filename + ".1")
    except:
        pass
    return

# ======================================================================
def write_toc(filename, lines):
    """ we always write utf8 """
    data = u"".join(lines)
    backup_file(filename)
    out = file(filename, "w")
    out.write(encode(data))
    out.close()

# ======================================================================
def write_index(filename, line):
    """ we always write utf8 """
    out = file(filename, "w")
    out.write(encode(line))
    out.close()

# ======================================================================
def get_genre(config):
    """
    get genre either from commandline arg or index file
    """
    if config['genre'] != "": return config['genre']

    try:
        tokens = read_index_file(config['file_index'])
        tokens = read_index_file(config['file_index'])
        return tokens[0]
    except Exception, err:
        logging.error('cannot read/parse index file (try -g option) ' + str(err))
        raise SyntaxError
    return
# ======================================================================
def get_id(config):
    """
    get id either from commandline arg or toc file
    """
    if config['id'] != "": return config['id']

    try:
        (dummy, discids) = read_toc_file(config['file_toc'])
        return discids[0]
    except Exception, err:
        logging.error('cannot read/parse toc file ' + str(err))
        raise SyntaxError
    return
# ======================================================================

_PATTERN_DISC_TRACK = re.compile(r'^# Track frame.*$')
_PATTERN_DISC_LENGTH = re.compile(r'^# Disc length.*$')


def merge_id_into_toc(config):

    def find_track(lines):
        for  i in range(len(lines)):
            if _PATTERN_DISC_TRACK.search(lines[i]):
                return i
        return -1

    def find_length(lines):
        for  i in range(len(lines)):
            if _PATTERN_DISC_LENGTH.search(lines[i]):
                return i
        return -1


    tracks = read_id_file(config['file_id'])
    length = tracks[-1]
    tracks = tracks[2:-1]

    (lines, dummy) = read_toc_file(config['file_toc'])

    i = find_track(lines)
    for tr in range(len(tracks)):
        newline  =  u"# " + tracks[tr] + u"\r\n"
        logging.info("Replacing >" + lines[i+tr][:-1] + "< with >" + newline[:-1] + "<")
        lines[i+tr] = newline

    i = find_length(lines)
    newline =  u"# Disc length: " + length + u"\r\n"
    logging.info("Replacing >" + lines[i][:-1] + "< with >" + newline[:-1] + "<")
    write_toc(config['file_toc'], lines)
    return

# ======================================================================
def make_command_url(config, command):
    """
    construct request url for freedb's http protocol
    """
    config['command'] = command
    return "%(server)s%(cgi_query)s?cmd=%(command)s&"\
           "hello=%(user)s+%(host)s+%(client)s+%(version)s&proto=%(protocol)s" % config

# ======================================================================
def cddb_motd(config):
    """
    execute the cddb motd (message of the day) command
    """
    url = make_command_url(config, "motd")

    response = urllib2.urlopen(url)

    out = response.read()
    print out
    return

# ======================================================================
def cddb_lscat(config):
    """
    execute the cddb lscat (list categories) command
    """

    url = make_command_url(config, "cddb+lscat")
    logging.info("query: [%s]", url)
    response = urllib2.urlopen(url)

    out = response.read()
    print out
    return

# ======================================================================

def cddb_query(config, tracks):
    """
    execute the cddb query command
    """

    result = []
    query_str = " ".join(tracks)

    #print query_str

    query_str = urllib.quote_plus(query_str.rstrip())

    url = make_command_url(config, "cddb+query+" + query_str)
    logging.info("query: [%s]", url)
    response = urllib2.urlopen(url)

    line = decode(response.readline(), config['encodings'])
    components = line.split()

    code = int(components[0])
    line = " ".join(components[1:])

    if code == 200:
        result.append(line)

    elif code == 202:
        pass

    elif code == 211 or code == 210:
        for line in response.readlines():
            if line[0] == '.': break
            result.append(decode(line[0:-1], config['encodings']))
    else:
        logging.warning("unknown return code %s\n" % components[0])

    return result

# ======================================================================

def cddb_read(config, genre, discid):
    """
    read a freedb record
    """
    url = make_command_url(config,"cddb+read+" + genre + "+" + discid )

    response = urllib2.urlopen(url)

    header = response.readline().rstrip()
    header = decode(header, config['encodings'])
    header = header.split(' ', 3)
    try:
        lines = [decode(x, config['encodings']) for x in response.readlines()]
    except UnicodeDecodeError:
        Error("encoding is not unicode try adding '-e' to command line")

    if lines == []:
        Error("no results returned from database\n")

    line = lines[-1]
    if line[0] != '.':
        Error("toc not terminated")

    return lines[0:-1]

# ======================================================================
def cddb_submit(config, toc, discid, genre):
    """
    execute cddb submit command
    """
    header = {}
    header['Category']  = genre
    header['Discid'] = discid
    header['User-Email'] = config['user'] + "@" + config['host']
    header['Submit-Mode'] = config['mode']
    header['Charset'] = 'UTF-8'
    header['X-Cddbd-Note'] = 'Sent by ' + config['client'] + " " + config['version']

    url = config['server'] + config['cgi_submit']
    #print url,header
    request = urllib2.Request(url, None, header)

    data = encode(u"".join(toc))
    #data = data[0:-3]

    response = urllib2.urlopen(request, data)

    for line in response.readlines():
        print decode(line, config['encodings'])
    return


def install_opener_with_redirect_handling():
    redirect_handler = urllib2.HTTPRedirectHandler()
    opener = urllib2.build_opener(redirect_handler)
    # ...and install it globally so it can be used with urlopen.
    urllib2.install_opener(opener)
    return

# ======================================================================

def main(argv):

    install_opener_with_redirect_handling()
    # DEFAULTS

    Config = {}
    Config['version'] = version()
    Config['server'] = 'http://gnudb.gnudb.org/'
    Config['protocol'] = '5'
    Config['user'] = getpass.getuser()
    Config['host'] = socket.gethostname()
#    Config['client'] = 'freedbtool'
#    Config['client'] = 'winamp 2.80'
    Config['client'] = 'xmcd'
    Config['mode'] = 'submit'
    Config['cgi_query'] = '~cddb/cddb.cgi'
    Config['cgi_submit'] = '~cddb/submit.cgi'
    Config['device'] = '/dev/cdrom'
    Config['file_id'] = 'id'
    Config['file_index'] = 'index'
    Config['file_toc'] = 'toc'
    Config['select'] = -1
    Config['encodings'] = ["utf8", "latin1"]   #internally we use unicode
    Config['genre'] = ""
    Config['id'] = ""

    # COMMAND LINE PARSING

    try:
        mode = argv[0]
        opts, dummy = getopt.getopt(argv[1:], "m:d:t:n:h:u:x:s:t:i:p:c:e:g:q:")
        for opt, arg in opts:
            if opt == '-n':
                Config['select'] =  int(arg)
                continue
            if opt == '-t':
                Config['file_toc'] =  arg
                continue
            if opt == '-x':
                Config['file_index'] =  arg
                continue
            if opt == '-i':
                Config['file_id'] =  arg
                continue
            if opt == '-h':
                Config['host'] =  arg
                continue
            if opt == '-u':
                Config['user'] =  arg
                continue
            if opt == '-s':
                Config['server'] =  arg
                continue
            if opt == '-d':
                Config['device'] =  arg
                continue
            if opt == '-m':
                Config['mode'] =  arg
                continue
            if opt == '-p':
                Config['protocol'] =  arg
                continue
            if opt == '-c':
                Config['client'] =  arg
                continue
            if opt == '-g':
                Config['genre'] =  arg
                continue
            if opt == '-q':
                Config['id'] =  arg
                continue
            if opt == '-e':
                Config['encodings'] =  arg.split(",")
                continue

    except Exception, err:
        logging.error("unknown option " + str(err))
        return usage()

    #
    # MODE DISPATCHING
    #

    if mode == 'track':
        merge_id_into_toc(Config)
    elif mode == 'motd':
        cddb_motd(Config)
    elif mode == 'lscat':
        cddb_lscat(Config)
    elif mode == 'config':
        print "Version $Revision: 806 $"
        for k in Config.keys():
            print k, Config[k]
    elif mode == 'submit':
        (lines, discids) = read_toc_file(Config['file_toc'])
        print lines
        genre = get_genre(Config)
        increment_revision(lines)
        replace_submiter(Config, lines)
        logging.warning("submitting: disk " + discids[0] + " " +
                        genre + " as " + Config['user'] + "@" + Config['host'])
        cddb_submit(Config, lines, discids[0], genre)

    elif mode == 'update':
        diskid = get_id(Config)
        genre = get_genre(Config)
        logging.info("updating %s %s" % (genre, diskid))
        toc = cddb_read(Config, genre, diskid)
        write_toc(Config['file_toc'], toc)
    elif mode == 'scan':
        diskid = get_id(Config)
        for genre in _CDDB_GENRES:
            try:
                logging.warning("scanning %s %s" % (genre, diskid))
                toc = cddb_read(Config, genre, diskid)
                print "".join(toc)
            except Exception:
                pass
    elif mode == 'get':
        if  Config['device']:
            ReadMetaData(Config['device'],  Config['file_id'])

        tracks = read_id_file( Config['file_id'] )
        results = cddb_query(Config, tracks)
        logging.info("found %d records" % len(results))
        num_selected = Config['select']
        num_results =  len(results)
        if num_results == 1:
            num_selected = 0

        if 0 <= num_selected and  num_selected < num_results:
            index = results[num_selected]
            logging.info("using number %d: %s" % (num_selected, index))
            write_index(Config['file_index'], index)
            part = index.split()
            toc = cddb_read(Config, part[0], part[1])
            write_toc(Config['file_toc'], toc)
        elif num_results == 0:
            logging.warning("no matches found")
        else:
            logging.warning("multiple matches found. "
                            "Use -n options to choose the right one.")
            for i in range(num_results):
                index = results[i]
                print "#" * 70
                print i, index
                print "#" * 70
                part = index.split()
                toc = cddb_read(Config, part[0], part[1])
                print encode(" ".join(toc))

    else:
        return usage()

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    sys.exit(main(sys.argv[1:]))

# eof



